from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Tuple, Union, TYPE_CHECKING
import sympy as sp
import numpy as np
from pyomo.environ import (
    ConcreteModel, RangeSet, Param, Var, Constraint, ConstraintList,
)
from sympy import Matrix as Mat
from . import utils, variable_list, visual
from . import collocation as _collocation
from .utils import flatten

if TYPE_CHECKING:
    from .links import Link3D

T = TypeVar('T')


def getattrs(items: Iterable, attr: str) -> List:
    """
    Get the attribute `attr` from every element in `items`, which is usually a list of links. Eg:
    >>> getattrs(self.links, 'mass')
    """
    return [getattr(item, attr) for item in items]


class System3D:
    # m: Union[ConcreteModel, None] = None
    def __init__(self, name: str, links: List['Link3D']) -> None:
        self.name = name
        self.links = links
        self.m: Union[ConcreteModel, None] = None

    def add_link(self, link: 'Link3D') -> None:
        self.links.append(link)

    def get_state_vars(self) -> Tuple[Mat, Mat, Mat]:
        q = flatten(getattrs(self.links, 'q'))
        dq = flatten(getattrs(self.links, 'dq'))
        ddq = flatten(getattrs(self.links, 'ddq'))
        return Mat(q), Mat(dq), Mat(ddq)

    def calc_eom(self, *, simp_func: Callable[[T], T] = lambda x: x) -> None:
        q, dq, ddq = self.get_state_vars()

        Ek, Ep, _, _ = utils.calc_velocities_and_energies(
            getattrs(self.links, 'Pb_I'),
            getattrs(self.links, 'Rb_I'),
            getattrs(self.links, 'mass_sym'),
            getattrs(self.links, 'inertia'),
            q, dq, g=9.81
        )
        Ek = simp_func(Ek)
        Ep = simp_func(Ep)

        M, C, G = utils.manipulator_equation(Ek, Ep, q, dq)
        M = simp_func(M)
        C = simp_func(C)
        G = simp_func(G)

        # get angular constraints
        angle_constraints = simp_func(Mat(
            flatten(link.angle_constraints for link in self.links)))
        Fr = Mat(flatten(link.constraint_forces for link in self.links))

        # foot stuff
        from . import foot
        feet = foot.feet(self)

        Q = sp.zeros(*q.shape)
        for link in self.links:
            Q += link.calc_eom(q, dq, ddq)

        B = simp_func(Q)

        from . import motor
        self.force_scale = sp.Symbol('F_{scale}')
        to_sub = {force: force*self.force_scale for force in [  # TODO: this should be done by the links themselves!
            *flatten(torque.input_torques for torque in motor.torques(self)),
            *Fr,
            *flatten(foot.Lx for foot in feet),
            *[foot.Lz for foot in feet],
        ]}

        eom = M @ ddq + C + G - B
        eom = simp_func(Mat([*eom, *angle_constraints]).xreplace(to_sub))

        # eom_c = M @ ddq + G - B
        # eom_c = simp_func(Mat([*eom_c, *angle_constraints]).xreplace(to_sub))

        self.sp_variables: List[sp.Symbol] = flatten(
            link.get_sympy_vars() for link in self.links
        )

        utils.info(f'Number of operations in EOM is {sp.count_ops(eom)}')

        # TODO: the lambdifying step actually takes quite long -- any way to speed it up?
        from pyomo.environ import atan
        func_map = {
            'sqrt': lambda x: (x+1e-9)**(1/2),
            'atan': atan,
            'atan2': lambda y, x: 2 * atan(y/((x**2 + y**2 + 1e-9)**(1/2) + x))
        }
        self.eom_f = utils.lambdify_EOM(
            eom,
            self.sp_variables + [self.force_scale],
            func_map=func_map
        )

    def make_pyomo_model(self, nfe: int, collocation: str, total_time: float,
                         scale_forces_by: Optional[float] = None,
                         vary_timestep_within: Optional[Tuple[float, float]] = None,
                         presolve_no_C: bool = False,
                         include_dynamics: bool = True) -> None:
        """
        vary_timestep_within:
            the upper and lower bounds for how much the unscaled timestep can change.
            Eg: given a timestep of t=50ms and vary_timestep_within=(0.8, 1.2),
            the timestep can vary from 40ms to 60ms

        presolve_no_C:
            whether the model should be pre-solved without the centrifugal forces
        """
        _collocation.check_collocation_method(collocation)

        if presolve_no_C:
            raise NotImplementedError(
                'Both EOM and EOM no C are saved to the model, but EOM no C is not used')

        self.m = m = ConcreteModel(name=self.name)

        # time and collocation
        ncp = _collocation.get_ncp(collocation)
        m.fe = RangeSet(nfe)
        m.cp = RangeSet(ncp, name=collocation)
        m.hm0 = Param(initialize=total_time/nfe)

        if vary_timestep_within is not None:
            m.hm = Var(m.fe, initialize=1.0, bounds=vary_timestep_within)
        else:
            m.hm = Param(m.fe, initialize=1.0)

        # add constraints using pyomo_model.constraints.add( Constraint )
        m.constraints = ConstraintList()

        # 1: each link/body defines its own pyomo variables
        for link in self.links:
            link.add_vars_to_pyomo_model(m)

        # 2: the variables are all combined
        self.pyo_variables = variable_list.VariableList(m, self.links)

        # 3: the equations of motion are all added
        for link in self.links:
            link.add_equations_to_pyomo_model(
                self.sp_variables, self.pyo_variables, collocation)

        if include_dynamics is True:
            if scale_forces_by is None:
                total_mass = sum(link.mass for link in self.links)
                scale_forces_by = total_mass
            m.force_scale = Param(initialize=scale_forces_by)

            @m.Constraint(m.fe, m.cp, range(len(self.eom_f)))
            def EOM_constr(m, fe, cp, i):
                return self.eom_f[i]([*self.pyo_variables[fe, cp], m.force_scale]) == 0 \
                    if not (fe == 1 and cp < ncp) else Constraint.Skip
        else:
            utils.warn('Not including dynamics (EOM_constr) in model')

    def save_data_to_dict(self, description: str) -> Dict[str, Any]:
        return {
            'name': self.name,
            'description': description,
            'repr': self.__repr__(),
            'nfe': len(self.m.fe),
            'ncp': len(self.m.cp),
            'hm': utils.get_vals(self.m.hm),
            'hm0': float(self.m.hm0.value),
            'links': [link.save_data_to_dict() for link in self.links],
        }

    def save_data_to_file(self, filename: str, description: str, overwrite_existing: bool = False) -> None:
        import dill
        import os.path
        if os.path.isfile(filename) and overwrite_existing is False:
            raise FileExistsError

        with open(filename, 'wb') as f:
            dill.dump(self.save_data_to_dict(description), f, recurse=True)

    def init_from_dict_one_point(self, data: Dict[str, Any], fed: int, cpd: int, fes: Optional[int] = None, cps: Optional[int] = None, **kwargs):
        """fed and cpd are destination (ie pyomo, and 1-indexed), fes and cps are source (ie dictionary, and 0-indexed)"""
        assert self.m is not None

        if fes is None:
            fes = fed - 1
        if cps is None:
            cps = cpd - 1

        # TODO: ignore this warning when accounting for differing number of finite elements
        if not self.m.hm0.value == data['hm0']:
            utils.warn(
                f'init_from_dict_one_point: self.hm0 = {self.m.hm0.value} != {data["hm0"]} = data["hm0"]', once=True)

        if utils.has_variable_timestep(self.m):
            utils.maybe_set_var(self.m.hm[fed], data['hm'][fes], **kwargs)

        for link, linkdata in zip(self.links, data['links']):
            link.init_from_dict_one_point(
                linkdata, fed, cpd, fes, cps, **kwargs)

    def init_from_dict(self, data: Dict[str, Any], **kwargs):
        for fed, cpd in self.indices(one_based=True):
            self.init_from_dict_one_point(data, fed, cpd, **kwargs)

    def init_from_file(self, filename: str, **kwargs):
        import dill
        # this could just be:
        # >>> self.init_from_dict(dill.load(open(filename, 'rb')), **kwargs)
        with open(filename, 'rb') as f:
            data = dill.load(f)

        self.init_from_dict(data, **kwargs)

    def indices(self, *, one_based: bool, skipfirst: bool = True) -> List[Tuple[int, int]]:
        return utils.get_indexes(len(self.m.fe), len(self.m.cp), one_based=one_based, skipfirst=skipfirst)

    def __getitem__(self, linkname: str) -> 'Link3D':
        for link in self.links:
            if link.name == linkname:
                return link

        raise KeyError(
            f'No link with name "{linkname}". Available links are: {", ".join(link.name for link in self.links)}')

    def plot_keyframes(self,
                       keyframes: List[int],
                       view_along: Union[Tuple[float, float], str],
                       plot3d_config: Dict = {},
                       save_to: Optional[str] = None,
                       ground_color: str = 'lightgray',
                       lims: Optional[Tuple[Tuple, Tuple, Tuple]] = None):
        # need to import this to get 3D plots working, for some reason
        from mpl_toolkits import mplot3d

        # typing this as 'Any' because the method accesses following give
        # false warnings otherwise...
        fig, ax, add_ground = visual.plot3d_setup(
            scale_plot_size=False, **plot3d_config)
        if lims is not None:
            x, y, z = lims
            ax.set_xlim(*x)
            ax.set_ylim(*y)
            ax.set_zlim(*z)

        visual.set_view(ax, along=view_along)

        ncp = len(self.m.cp)

        cp = ncp
        data: List[List[float]] = [
            [v.value for v in self.pyo_variables[fe, cp]]
            for fe in self.m.fe
        ]

        try:
            add_ground((ax.get_xlim(), ax.get_ylim()), color=ground_color)
            for fe in keyframes:
                # iterate in reverse as matplotlib plots that last things on top, and we want
                # the body on top. Luckily, this also sorts out left vs right for the 3D
                # quadruped model, but it's a bad hack in general :/
                # a better approach would be to sort the order of the links such that those
                # furthest from the camera are plotted first
                # the best would obviously be if matplotlib's 3D plotting was improved, or if
                # we switch libraries
                for link in self.links[::-1]:
                    link.animation_setup(fig, ax, data)
                    link.animation_update(fig, ax, fe=fe, track=False)

            if save_to is not None:
                fig.savefig(save_to)

        except Exception as e:
            visual.error(f'Interrupted keyframes due to error: {e}')

        finally:
            for link in self.links:
                link.cleanup_animation(fig, ax)
            del fig, ax

    def animate(self,
                view_along: Union[Tuple[float, float], str],
                t_scale: float = 1.,
                camera: Optional[Union[Tuple[float, float],
                                       Tuple[float, float, float]]] = None,
                lim: Optional[float] = None,
                plot3d_config: Dict = {},
                lims: Optional[Tuple[Tuple, Tuple, Tuple]] = None,
                track: Optional[str] = None,
                dt: Optional[float] = None,
                save_to: Optional[str] = None):
        # need to import this to get 3D plots working, for some reason
        from mpl_toolkits import mplot3d
        import matplotlib.animation
        import matplotlib.pyplot as plt

        assert self.m is not None, \
            'robot does not have a pyomo model defined on it'

        # typing this as 'Any' because the method accesses following give
        # false warnings otherwise...
        fig, ax, add_ground = visual.plot3d_setup(
            scale_plot_size=False, **plot3d_config)

        if lims is not None:
            x, y, z = lims
            ax: Any
            ax.set_xlim(*x)
            ax.set_ylim(*y)
            ax.set_zlim(*z)

        visual.set_view(ax, along=view_along)
        ground = add_ground(((0, 0), (0, 0)))

        nfe = len(self.m.fe)
        ncp = len(self.m.cp)

        cp = ncp
        data: List[List[float]] = [
            [v.value for v in self.pyo_variables[fe, cp]]
            for fe in self.m.fe
        ]

        for link in self.links:
            link.animation_setup(fig, ax, data)

        if camera is not None and lim is not None:
            visual.track_pt(ax, camera, lim)

        def progress_bar(proportion: float, width: int = 80):
            import sys
            # a mini progress bar
            num_done = int(round(proportion * width))
            if num_done == width:
                sys.stdout.write(' '*width + '\r')
            else:
                sys.stdout.write('+'*num_done + '-'*(width-num_done) + '\r')

        if dt is None:
            def _animate(fe: int):  # fe is one-based
                progress_bar(((fe-1)*ncp + cp)/(nfe*ncp))

                for link in self.links:
                    link.animation_update(
                        fig, ax, fe=fe, track=(track == link.name))

                nonlocal ground
                ground.remove()
                ground = add_ground((ax.get_xlim(), ax.get_ylim()))
        else:
            if not utils.has_variable_timestep(self.m):
                t_arr = np.cumsum(
                    np.array([self.m.hm[fe] for fe in self.m.fe]))
            else:  # the t_arr below is variable step
                t_arr = np.cumsum(utils.get_vals(self.m.hm)) * self.m.hm0.value

            def _animate(t: float):  # t is a float from 0 to total_time
                progress_bar(t/frames[-1])

                for link in self.links:
                    link.animation_update(
                        fig, ax, t=t, t_arr=t_arr, track=(track == link.name))

                nonlocal ground
                ground.remove()
                ground = add_ground((ax.get_xlim(), ax.get_ylim()))

        # if you get impatient and cancel an animation while it's being made,
        # then try to clone a model, you get an error from matplotlib about
        # certain things not being cloneable. So -> try/catch/finally block
        try:
            if not utils.has_variable_timestep(self.m):
                t_sum = sum(self.m.hm[fe]
                            for fe in self.m.fe if fe != nfe) * self.m.hm0.value
            else:
                t_sum = sum(
                    self.m.hm[fe].value for fe in self.m.fe if fe != nfe) * self.m.hm0.value

            interval_ms = 1000*t_scale * \
                (t_sum / nfe / ncp if dt is None else dt)
            frames = [fe for fe in self.m.fe] if dt is None else np.arange(
                start=0, stop=t_sum+dt, step=dt)
            anim = matplotlib.animation.FuncAnimation(
                fig, _animate, frames=frames, interval=interval_ms
            )

            plt.close(anim._fig)

            if save_to is not None:
                anim.save(save_to)
            elif utils.in_ipython():
                from IPython.core.display import display, HTML
                display(HTML(anim.to_html5_video()))
            else:
                fig.show()

        except Exception as e:
            progress_bar(nfe, ncp)
            visual.error(f'Interrupted animation due to error: {e}')

        finally:
            for link in self.links:
                link.cleanup_animation(fig, ax)
            del fig, ax

    def plot(self, plot_links: bool = True) -> None:
        assert self.m is not None
        
        if utils.has_variable_timestep(self.m):
            import matplotlib.pyplot as plt
            plt.title('Timestep size vs finite element')
            data = 1000*utils.get_vals(self.m.hm) * self.m.hm0.value
            plt.plot(data)
            plt.ylabel('Timestep size [ms]')
            plt.ylim([0, max(data)*1.1])
            plt.show()

        if plot_links:
            for link in self.links:
                link.plot()

    def __repr__(self) -> str:
        child_links = '\n  '.join(str(link) + ',' for link in self.links)
        return f'System3D(name="{self.name}", [\n  {child_links}\n])'

    def post_solve(self, costs: Optional[Dict[str, Any]] = None, detailed: bool = False, tol: float = 1e-6):
        from .foot import feet_penalty
        from pyomo.environ import value as pyovalue
        print('Total cost:', pyovalue(self.m.cost))

        if costs is not None:
            for k, v in costs.items():
                print(f'-- {k}: {pyovalue(v)}')

        foot_pen = pyovalue(feet_penalty(self))
        if foot_pen > 1e-3:
            visual.error('Foot penalty seems to be unsolved')

        if detailed is True:
            from pyomo.util.infeasible import log_infeasible_constraints
            print('Infeasible constraints:')
            log_infeasible_constraints(self.m, tol=tol)

            from pyomo.util.infeasible import log_infeasible_bounds
            print('Infeasible bounds:')
            log_infeasible_bounds(self.m, tol=tol)

    # def init_from_robot(self, source: 'System3D', interpolation: str = 'linear'):
    #     """
    #     Initialize a model from a solved one, interpolating values if needed
    #     """
    #     # TODO: change this to be more like the dict approach above! Or just use that instead!
    #     utils.debug('Only use init_from_robot with two robots which are identical, but'
    #                 ' with possibly differing finite elements and collocation points')
    #     utils.warn('init_from_robot: haven\'t yet figured out what to do with `None`s at'
    #                ' the beginning of things')
    #     utils.warn('init_from_robot: some things (like dummy variables) aren\'t yet'
    #                ' copied across')
    #     # attempt to catch errors early on. TODO: think of more quick tests
    #     assert len(self.links) == len(source.links)
    #     assert interpolation == 'linear'

    #     import math

    #     def valid_num(num): return not (
    #         num is None or math.isnan(num) or math.isinf(num)
    #     )

    #     # TODO: refactor into a different func?
    #     # def supersample(new_x, old_x, data):
    #     #     return np.interp(
    #     #         np.linspace(0, 1, num=new_x),
    #     #         np.linspace(0, 1, num=old_x),
    #     #         data
    #     #     )

    #     # get data from source
    #     hm = utils.get_vals(source.m.hm)

    #     # interpolate
    #     hm2 = np.interp(
    #         np.linspace(0, 1, num=len(self.m.fe)),
    #         np.linspace(0, 1, num=len(source.m.fe)),
    #         hm)

    #     # add to destination model
    #     for fe in self.m.fe:
    #         self.m.hm[fe].value = hm2[fe-1] if valid_num(hm2[fe-1]) else 1.0

    #     for destlink, srclink in zip(self.links, source.links):
    #         # get data from source. using `astype` to get rid of None's
    #         srcdata = np.array([
    #             var.value for fe in source.m.fe
    #             for cp in source.m.cp
    #             for var in srclink.get_pyomo_vars(fe, cp)
    #         ]).reshape((len(source.m.fe) * len(source.m.cp), -1)).astype(float)

    #         # x interpolation points (replace with cumsum of hm!)
    #         x_orig = np.linspace(0, 1, num=len(source.m.fe) * len(source.m.cp))
    #         x_dest = np.linspace(0, 1, num=len(self.m.fe) * len(self.m.cp))

    #         destdata = [[destlink.get_pyomo_vars(fe, cp) for cp in self.m.cp]
    #                     for fe in self.m.fe]

    #         # interpolate
    #         interpolated_data = np.zeros(
    #             (x_dest.shape[0], srcdata.shape[1]))  # type: ignore
    #         for varidx in range(srcdata.shape[1]):
    #             interped = np.interp(x_dest, x_orig, srcdata[:, varidx])
    #             # replace the np.nan's with None, for pyomo
    #             interpolated_data[:, varidx] = np.where(
    #                 np.isnan(interped), None, interped)  # type: ignore

    #         # add to destination model
    #         ncp = len(self.m.cp)
    #         skipped_vars = []
    #         for fe, cp in self.indices(one_based=False):
    #             for varidx, var in enumerate(destdata[fe][cp]):
    #                 num = interpolated_data[fe*ncp + cp, varidx]

    #                 if var.is_fixed():
    #                     skipped_vars.append(var)
    #                     continue
    #                 if not valid_num(num):
    #                     print('skipping invalid number:', num,
    #                           'at index', varidx, 'for variable', var)
    #                     continue

    #                 var.value = num

    #         if len(skipped_vars) > 0:
    #             from textwrap import shorten
    #             utils.debug(
    #                 shorten(f'skipped variables because they are fixed: {skipped_vars}', width=100))


    # # Figure out how to handle constraints etc with this presolve approach. Eg min distance, etc
    # def presolve(self, collocation: str, nfe: int, setup_func: Callable[['System3D'], None], no_C: bool,
    #              make_pyomo_model_kwargs: dict = {}, default_solver_kwargs: dict = {}):
    #     """
    #     Create a new (simpler) model, solve it, then copy over the (interpolated) values
    #     to this model. Note -- this function is definitely worth reading through in detail, especially
    #     to know what defaults it chooses for you! If you're not sure, rather don't use this!

    #     Example:
    #     >>> robot.make_pyomo_model(nfe=50, collocation='radau', total_time=1.5)
    #     >>> def add_task(robot):
    #     ...     add_pyomo_constraints(robot)
    #     ...     high_speed_stop(robot, initial_vel=20.0)
    #     >>> add_task(robot)
    #     >>> solver_kwargs = {'OF_print_frequency_time': 10, 'OF_hessian_approximation': 'limited-memory', }
    #     >>> robot.presolve(nfe=10, collocation='euler', setup_func=add_task, no_C=True,
    #     ...                default_solver_kwargs=solver_kwargs)
    #     >>> ret = utils.default_solver(max_mins=30, solver='ma86', **solver_kwargs).solve(robot.m, tee=True)
    #     """
    #     import copy

    #     new_sys = copy.deepcopy(self)
    #     new_sys.make_pyomo_model(nfe=nfe, collocation=collocation, presolve_no_C=no_C,
    #                              total_time=float(
    #                                  self.m.hm0.value * len(self.m.fe)),
    #                              **make_pyomo_model_kwargs)
    #     setup_func(new_sys)

    #     results = utils.default_solver(max_mins=10, solver='ma86',
    #                                    OF_hessian_approximation='limited-memory',
    #                                    **default_solver_kwargs).solve(new_sys.m, tee=True)

    #     from pyomo.opt import TerminationCondition
    #     if results.solver.termination_condition == TerminationCondition.infeasible:  # type: ignore
    #         utils.warn('Presolving returned an infeasible result')

    #     self.init_from_robot(new_sys)



from typing import Any, Dict, List, Callable, Optional, Tuple, Union, Iterable, TYPE_CHECKING
from typing_extensions import TypedDict
import sympy as sp
import numpy as np

from pyomo.environ import (
    ConcreteModel, Param, Set, Var, Constraint,
)
from sympy import Matrix as Mat
from .system import System3D
from .links import Link3D
from . import utils


if TYPE_CHECKING:
    from .variable_list import VariableList

PlotConfig = TypedDict('PlotConfig', plot_forces=bool, force_scale=float)


class Foot3D:
    def __init__(self, name: str, Pb_I: Mat, nsides: int, friction_coeff: Optional[float] = None):
        self.name = name
        self.nsides = nsides
        self.Pb_I = Pb_I
        self.friction_coeff = friction_coeff

        # the contact/friction stuff:
        self.D = friction_polygon(nsides)
        self.Lx = Mat(sp.symbols('L_{%s/x:%s}' % (name, nsides)))
        self.Lz = sp.Symbol('L_{%s/z}' % name)
        self.L = Mat(self.D).T @ self.Lx + Mat([0, 0, self.Lz])

        self._plot_config: PlotConfig = {
            'plot_forces': True,
            'force_scale': 1/10,
        }

    def calc_eom(self, q, dq, ddq) -> Mat:
        self.Pb_I_vel = self.Pb_I.jacobian(q) @ dq

        jac_L = self.Pb_I.jacobian(q)
        return jac_L.T @ self.L

    def add_vars_to_pyomo_model(self, m: ConcreteModel):
        assert isinstance(self.friction_coeff, float), \
            f'The friction_coeff for {self.name} must be set to a float'

        # parameter and sets
        friction_coeff = Param(initialize=self.friction_coeff,
                               name='friction_coeff')
        xy_set = Set(initialize=('x', 'y'), name='xy_set', ordered=True)
        fric_set = Set(initialize=range(8), name='fric_set', ordered=True)

        GRFxy = Var(m.fe, m.cp, fric_set, name='GRFxy', bounds=(0, 30))
        GRFz = Var(m.fe, m.cp,            name='GRFz',  bounds=(0, 30))

        # dummy vars equal to parts from EOM
        foot_height = Var(m.fe, m.cp, name='foot_height', bounds=(0, None))
        foot_xy_vel = Var(m.fe, m.cp, xy_set, name='foot_xyvel')
        gamma = Var(m.fe, m.cp, name='gamma (foot xy-velocity magnitude)',
                    bounds=(0, None))

        # penalty variables
        contact_penalty = Var(m.fe, name='contact_penalty', bounds=(0, 10))
        friction_penalty = Var(m.fe, name='friction_penalty', bounds=(0, 10))
        slip_penalty = Var(m.fe, fric_set, name='slip_penalty', bounds=(0, 10))

        self.pyomo_params: Dict[str, Param] = {
            'friction_coeff': friction_coeff
        }
        self.pyomo_sets: Dict[str, Set] = {
            'xy_set': xy_set, 'fric_set': fric_set,
        }
        self.pyomo_vars: Dict[str, Var] = {
            'GRFxy': GRFxy, 'GRFz': GRFz,
            'foot_height': foot_height, 'foot_xy_vel': foot_xy_vel, 'gamma': gamma,
            'contact_penalty': contact_penalty,
            'friction_penalty': friction_penalty,
            'slip_penalty': slip_penalty,
        }

        utils.add_to_pyomo_model(m, self.name, [
            self.pyomo_params.values(),
            self.pyomo_sets.values(),
            self.pyomo_vars.values(),
        ])

    def get_pyomo_vars(self, fe: int, cp: int):
        """fe, cp are one-based!"""
        # NB: keep in sync with get_sympy_vars()!!
        v = self.pyomo_vars
        return [*v['GRFxy'][fe, cp, :], v['GRFz'][fe, cp]]

    def get_sympy_vars(self):
        # NB: keep in sync with get_pyomo_vars()!!
        return [*self.Lx, self.Lz]

    def save_data_to_dict(self) -> Dict[str, Any]:
        fric_set = self.pyomo_sets['fric_set']

        return {
            'name': self.name,
            'nsides': self.nsides,
            'friction_coeff': self.friction_coeff,
            'contact_penalty': utils.get_vals(self.pyomo_vars['contact_penalty']),
            'friction_penalty': utils.get_vals(self.pyomo_vars['friction_penalty']),
            'slip_penalty': utils.get_vals(self.pyomo_vars['slip_penalty'], (fric_set,)),
            'foot_height': utils.get_vals(self.pyomo_vars['foot_height'], tuple()),
            'GRFz': utils.get_vals(self.pyomo_vars['GRFz'], tuple()),
            'GRFxy': utils.get_vals(self.pyomo_vars['GRFxy'], (fric_set,)),
        }

    def init_from_dict_one_point(self, data: Dict[str, Any], fed: int, cpd: int, fes: Optional[int] = None, cps: Optional[int] = None, **kwargs) -> None:
        if fes is None:
            fes = fed - 1
        if cps is None:
            cps = cpd - 1

        assert self.name == data['name']
        for attr in ('nsides', 'friction_coeff'):
            if getattr(self, attr) != data[attr]:
                utils.warn(
                    f'Attribute "{attr}" of link "{self.name}" is not the same as the data: {getattr(self, attr)} != {data[attr]}')

        v = self.pyomo_vars
        utils.maybe_set_var(v['contact_penalty'][fed],
                            data['contact_penalty'][fes], **kwargs)
        utils.maybe_set_var(v['friction_penalty'][fed],
                            data['friction_penalty'][fes], **kwargs)

        utils.maybe_set_var(v['foot_height'][fed, cpd],
                            data['foot_height'][fes, cps], **kwargs)
        utils.maybe_set_var(v['GRFz'][fed, cpd],
                            data['GRFz'][fes, cps], **kwargs)

        for (idx, f) in enumerate(self.pyomo_sets['fric_set']):
            utils.maybe_set_var(v['slip_penalty'][fed, f],
                                data['slip_penalty'][fes, idx], **kwargs)
            utils.maybe_set_var(v['GRFxy'][fed, cpd, f],
                                data['GRFxy'][fes, cps, idx], **kwargs)

    def add_equations_to_pyomo_model(self,
                                     sp_variables: List[sp.Symbol],
                                     pyo_variables: 'VariableList',
                                     collocation: str):
        friction_coeff = self.pyomo_params['friction_coeff']
        m = friction_coeff.model()

        xy_set = self.pyomo_sets['xy_set']  # foot_xy_vel below
        fric_set = self.pyomo_sets['fric_set']

        GRFxy = self.pyomo_vars['GRFxy']
        GRFz = self.pyomo_vars['GRFz']

        foot_height = self.pyomo_vars['foot_height']
        foot_xy_vel = self.pyomo_vars['foot_xy_vel']
        gamma = self.pyomo_vars['gamma']

        contact_penalty = self.pyomo_vars['contact_penalty']
        friction_penalty = self.pyomo_vars['friction_penalty']
        slip_penalty = self.pyomo_vars['slip_penalty']

        self.foot_pos_func = utils.lambdify_EOM(self.Pb_I, sp_variables)
        self.foot_xy_vel_func = utils.lambdify_EOM(
            self.Pb_I_vel[:2], sp_variables)

        ncp = len(m.cp)

        def add_constraints(name: str, func: Callable, indexes: tuple):
            setattr(m, self.name + '_' + name,
                    Constraint(*indexes, rule=func))

        def def_foot_height(m, fe, cp):    # foot height above z == 0 (xy-plane)
            if (fe == 1 and cp < ncp):
                return Constraint.Skip
            return foot_height[fe, cp] == self.foot_pos_func[2](pyo_variables[fe, cp])

        add_constraints('foot_height_constr', def_foot_height, (m.fe, m.cp))

        def def_foot_xy_vel(m, fe, cp, xy):  # foot velocity in xy-plane
            if (fe == 1 and cp < ncp):
                return Constraint.Skip
            i = 0 if xy == 'x' else 1
            return foot_xy_vel[fe, cp, xy] == self.foot_xy_vel_func[i](pyo_variables[fe, cp])

        add_constraints('foot_xy_vel_constr',
                        def_foot_xy_vel, (m.fe, m.cp, xy_set))

        def def_gamma(m, fe, cp, i):  # this sets gamma to the biggest of vx + vy
            if (fe == 1 and cp < ncp):
                return Constraint.Skip
            vx, vy = foot_xy_vel[fe, cp, 'x'], foot_xy_vel[fe, cp, 'y']
            return gamma[fe, cp] >= vx * self.D[i, 0] + vy * self.D[i, 1]

        add_constraints('gamma_constr', def_gamma, (m.fe, m.cp, fric_set))

        def def_friction_polyhedron(m, fe, cp):
            if (fe == 1 and cp < ncp):
                return Constraint.Skip
            return friction_coeff * GRFz[fe, cp] >= sum(GRFxy[fe, cp, :])

        add_constraints('friction_polyhedron_constr',
                        def_friction_polyhedron, (m.fe, m.cp))

        # complementarity equations
        # z[i+1]*GRFz[i] ≈ 0
        def def_contact_complementarity(m, fe):
            if fe < m.fe[-1]:
                α = sum(foot_height[fe+1, :])
                β = sum(GRFz[fe, :])
                return α * β <= contact_penalty[fe]
            else:
                return Constraint.Skip

        add_constraints('contact_complementarity_constr',
                        def_contact_complementarity, (m.fe,))

        # (μ * GRFz - Σ GRFxy) * γ ≈ 0
        def def_friction_complementarity(m, fe):
            α = friction_coeff * sum(GRFz[fe, :]) - sum(GRFxy[fe, :, :])
            β = sum(gamma[fe, :])
            return α * β <= friction_penalty[fe]

        add_constraints('friction_complementarity_constr',
                        def_friction_complementarity, (m.fe,))

        # GRFxy * (γ + dxyᵀ*Dᵢ) ≈ 0
        def def_slip_complementarity(m, fe, i):
            vx, vy = foot_xy_vel[fe, :, 'x'], foot_xy_vel[fe, :, 'y']
            α = sum(GRFxy[fe, :, i])
            β = sum(gamma[fe, :]) + sum(vx)*self.D[i, 0] + sum(vy)*self.D[i, 1]
            return α * β <= slip_penalty[fe, i]

        add_constraints('slip_complementarity_constr',
                        def_slip_complementarity, (m.fe, fric_set))

    def __getitem__(self, varname: str) -> Var:
        return self.pyomo_vars[varname]

    def penalty_sum(self):
        contact_penalty = self.pyomo_vars['contact_penalty']
        friction_penalty = self.pyomo_vars['friction_penalty']
        slip_penalty = self.pyomo_vars['slip_penalty']
        return (sum(contact_penalty[:])
                + 0.1*sum(friction_penalty[:])
                + 0.1*sum(slip_penalty[:, :]))

    def plot_config(self, *, plot_forces: Optional[bool] = None,
                    force_scale: Optional[float] = None) -> 'Foot3D':
        """Configuration for how this link should be plotted"""
        if plot_forces is not None:
            self._plot_config['plot_forces'] = plot_forces

        if force_scale is not None:
            self._plot_config['force_scale'] = force_scale

        return self

    def animation_setup(self, fig, ax, data: List[List[float]]):
        if self._plot_config['plot_forces'] is False:
            return

        self.has_line = False
        self.plot_data = np.empty((len(data), 6))
        cp = 1
        force_scale = self._plot_config['force_scale']

        for fe0, d in enumerate(data):
            fe = fe0 + 1

            x, y, z = [f(d) for f in self.foot_pos_func]

            dx, dy = 0, 0
            for f in self.pyomo_sets['fric_set']:
                dx, dy = [dx, dy] + (self['GRFxy'][fe, cp, f].value
                                     * self.D[f, :2] * force_scale)

            dz = self.pyomo_vars['GRFz'][fe, cp].value * force_scale

            self.plot_data[fe0, :] = (x, y, z, dx, dy, dz)

    def animation_update(self, fig, ax,
                         fe: Optional[int] = None,
                         t: Optional[float] = None,
                         t_arr: Optional[np.ndarray] = None,
                         track: bool = False):
        if self._plot_config['plot_forces'] is False:
            return

        if self.has_line:
            self.line.remove()
            self.has_line = False

        if fe is not None:
            x, y, z, dx, dy, dz = self.plot_data[fe-1]
        else:
            assert t is not None and t_arr is not None
            x, y, z, dx, dy, dz = [
                np.interp(t, t_arr, self.plot_data[:, i]) for i in range(6)]

        self.line = ax.quiver(
            x, y, z,    # <-- starting point of vector
            dx, dy, dz,  # <-- directions of vector
            arrow_length_ratio=0.15,
            color='red', alpha=.8, lw=1.5,
        )
        self.has_line = True

    def cleanup_animation(self, fig, ax):
        try:
            del self.line
        except:
            pass

    def plot(self):
        m = self.pyomo_vars['GRFxy'].model()
        fe = list(range(len(m.fe)))

        # xy_set  = self.pyomo_sets['xy_set']  # foot_xy_vel below

        fric_set = self.pyomo_sets['fric_set']
        contact_penalty = utils.get_vals(self.pyomo_vars['contact_penalty'])
        friction_penalty = utils.get_vals(self.pyomo_vars['friction_penalty'])
        slip_penalty = utils.get_vals(
            self.pyomo_vars['slip_penalty'], (fric_set,))

        # TODO: plot GRFxy as an xy-thing? or individual line plots?
        # GRFxy = utils.get_vals(self.pyomo_vars['GRFxy'], (fric_set,))

        # foot_xy_vel = self.pyomo_vars['foot_xy_vel']
        # gamma = self.pyomo_vars['gamma']

        import matplotlib.pyplot as plt

        plt.plot(fe, contact_penalty,  label='contact')
        plt.plot(fe, friction_penalty, label='friction')
        for fric in fric_set:
            plt.plot(fe, slip_penalty[:, fric], label=f'slip_{fric}')

        plt.title(f'Penalties in foot {self.name}')
        plt.legend()
        plt.grid(True)
        plt.show()

        fig, ax1 = plt.subplots()
        plt.title('Foot height and ground reaction force in ' + self.name)
        plt.grid(True)

        foot_height = utils.get_vals(self.pyomo_vars['foot_height'], tuple())
        ax1.plot(fe, foot_height, label='Foot height')
        ax1.set_xlabel('Finite element')
        ax1.set_ylabel('Height [m]')

        ax2 = ax1.twinx()
        GRFz = utils.get_vals(self.pyomo_vars['GRFz'], tuple())
        # the color trick below is so that they don't both use the same color
        ax2.plot(fe, GRFz, label='$GRFz$', color=next(
            ax1._get_lines.prop_cycler)['color'])
        ax2.set_ylabel('Force [Nm/body_weight]')

        fig.legend(loc='center')
        fig.tight_layout()  # otherwise the right y-label is slightly clipped
        plt.show()

    def __repr__(self) -> str:
        return f'Foot3D(name="{self.name}", nsides={self.nsides}, friction_coeff={self.friction_coeff})'


def add_foot(link, at: str, name: Optional[str] = None, **kwargs):
    """
        `at` is usually `self.bottom_I`. Eg:
        >>> foot.add_foot(link, at='bottom')
        """
    if name is None:
        name = link.name + '_foot'

    assert name not in link.nodes,\
        f'This link already has a node with the name {name}'

    assert at in ('top', 'bottom'), \
        "Can only add ground contacts at top or bottom of foot"

    Pb_I = link.bottom_I if at == 'bottom' else link.top_I
    foot = Foot3D(str(name), Pb_I, **kwargs)
    link.nodes[name] = foot
    return foot


def feet(robot_or_link: Union[System3D, Link3D]) -> List[Foot3D]:
    from typing import cast
    if isinstance(robot_or_link, System3D):
        robot = robot_or_link
        return [cast(Foot3D, node)
                for link in robot.links
                for node in link.nodes.values()
                if isinstance(node, Foot3D)]
    else:
        link = robot_or_link
        return [cast(Foot3D, node)
                for node in link.nodes.values()
                if isinstance(node, Foot3D)]


def feet_penalty(robot: 'System3D'):
    return sum(
        foot.penalty_sum()
        for foot in feet(robot)
    )


def friction_polygon(nsides: int) -> np.ndarray:
    if nsides == 4:
        # the four edges of a square - a very rough approximation to a friction cone
        return np.array([
            1, 0, 0,
            0, 1, 0,
            -1, 0, 0,
            0, -1, 0,
        ]).reshape(4, 3)

    elif nsides == 8:
        # for these numbers to make sense, think about moving around a square in the xy-plane,
        # going from an axis crossing, to an edge, to the next axis crossing, etc
        # this square gets normalized to approximate a unit circle
        D = np.array([
            1, 0, 0,
            1, 1, 0,
            0, 1, 0,
            -1, 1, 0,
            -1, 0, 0,
            -1, -1, 0,
            0, -1, 0,
            1, -1, 0,
        ]).reshape(8, 3)  # normalize by row
        return D / np.linalg.norm(D, axis=1).reshape(8, 1)  # type: ignore
    else:
        raise ValueError(
            'Only 4-sided and 8-sided friction polygons are implemented at the moment')


def interactively_set_timings(feet: Iterable[str], nfe: int, wait_until_all_set: bool = False, **kwargs) -> List:
    """
    >>> interactively_set_timings([link.foot.name for link in robot.links if link.has_foot()], nfe=50, time=0.35)
    """
    funcs = []
    for foot in feet:
        funcs.append(set_timing(description=foot, nfe=nfe, **kwargs))

    if wait_until_all_set is True:
        from time import sleep
        while not all(f()['done'] for f in funcs):
            sleep(0.5)
        return [f()['values'] for f in funcs]

    else:
        return funcs


def set_timing(nfe: int, *, time: Optional[float] = None, initial: Optional[Tuple] = None,
               description: str = 'Foot timing', width: str = '500px', when_done: str = 'disable') -> Callable[[], Dict[str, Union[bool, Tuple[int, int]]]]:
    """
    A simple example:
    >>> f = get_foot_timing(nfe=20, time=0.35)
    >>> from time import sleep
    >>> while not f()['done']: sleep(0.5)
    >>> print('The values are:', f()['values'])
    """
    import ipywidgets as widgets  # pyright: reportMissingImports=false

    assert when_done in ('close', 'disable')

    options = np.arange(nfe) + 1

    if time is not None:
        options = [(f'{o/nfe * time:.3}s', o) for o in options]
        # if initial is not None:
        #    initial = int(initial[0] * nfe/time), int(initial[1] * nfe/time)

    if initial is None:
        initial = (1, nfe)

    assert 1 <= initial[0] <= nfe and 1 <= initial[
        1] <= nfe, f'Invalid initial range: {initial}'
    slider = widgets.SelectionRangeSlider(
        value=initial,
        options=options,
        index=(0, len(options)-1),
        description=description,
        layout={'width': width}
    )

    button = widgets.Button(
        description='Click when done',
        disabled=False,
        button_style='',
        tooltip='Click me',
        icon='check'
    )

    done = False

    def finish_up(*args):
        nonlocal done
        done = True

        if when_done == 'close':
            slider.close()
            button.close()
        else:
            button.disabled = True
            slider.disabled = True

    button.on_click(finish_up)

    from IPython.display import display
    display(widgets.HBox([slider, button]))

    return lambda: {'done': done, 'values': slider.value}


def prescribe_contact_order(feet: List[Foot3D], ground_timings: Iterable[Tuple[int, int]],
                            min_foot_height: float = 0.005, min_GRFz: float = 0.001) -> None:
    """
    Prescribe `feet` to be on the ground using the ranges specified in `ground_timings`.
    `min_foot_height` sets the lower bound for feet during the flight phase, and `min_GRFz`
    does the same for the vertical ground reaction force during the stance phase.

    >>> feet = [link.foot for link in robot.links if link.has_foot()]
    >>> foot_order_vals = ((1, 9), (7, 15), (28, 37), (35, 43))  # corresponds to finite elements
    >>> list(zip(feet, foot_order_vals))
    [(Foot3D(name="LFL_foot", nsides=8), (1, 9)),
     (Foot3D(name="LFR_foot", nsides=8), (7, 15)),
     (Foot3D(name="LBL_foot", nsides=8), (28, 37)),
     (Foot3D(name="LBR_foot", nsides=8), (35, 43))]
    >>> prescribe_contact_order(feet, foot_order_vals)
    """
    def inclusive_range(start, stop): return range(start, stop+1)
    m = feet[0]['GRFz'].model()
    nfe = len(m.fe)

    def foot_fix_util(foot: Foot3D, start: int, stop: int):
        GRFz = foot['GRFz']
        foot_height = foot['foot_height']

        # phase 1: flight
        for fe in inclusive_range(1, start-1):
            for cp in m.cp:
                foot_height[fe, cp].setlb(min_foot_height)
                GRFz[fe, cp].fix(0)

        # phase 2: stance
        for fe in inclusive_range(start+1, stop-1):
            for cp in m.cp:
                foot_height[fe, cp].fix(0)
                GRFz[fe, cp].setlb(min_GRFz)

        # phase 3: flight
        for fe in inclusive_range(stop+1, nfe):
            for cp in m.cp:
                foot_height[fe, cp].setlb(min_foot_height)
                GRFz[fe, cp].fix(0)

    for foot, (start, stop) in zip(feet, ground_timings):
        foot_fix_util(foot, start, stop)



"""
    Number of operations in EOM without simplification: 186 903. With utils.parsimp: 

    Example usage:
    >>> import optim_lib; utils = optim_lib.utils
    >>> robot, add_pyomo_constraints = optim_lib.models.biped3D()
    >>> robot.calc_eom(simp_func = lambda x: utils.parsimp(x, nprocs=12))
    >>> robot.make_pyomo_model(nfe=30, collocation='euler', total_time=0.7, vary_timestep_within=(0.8, 1.2))
    >>> add_pyomo_constraints(robot)
    >>> costs = optim_lib.models.tasks.drop_test(robot, z_rot=0.2, min_torque=False, initial_height=1.0)
    >>> utils.default_solver(max_mins=10, solver='ma86', OF_hessian_approximation='limited-memory').solve(robot.m, tee=True)
    >>> robot.post_solve(costs)
    >>> robot.animate(view_along=(35, -120), t_scale=2, track='base')
"""

from typing import Callable, Tuple
from ..links import Link3D, constrain_rel_angle
from ..system import System3D
from sympy import Matrix as Mat
from ..foot import add_foot
from ..motor import add_torque


def model() -> Tuple[System3D, Callable[[System3D], None]]:
    body = Link3D('base', '+y', base=True)

    # for stabilization
    tail = Link3D('tail', '-x', start_I=body.Pb_I)
    body.add_hookes_joint(tail, about='xy')
    add_torque(body, tail, about='xy')

    # upper left, lower left
    L_thigh = Link3D('UL', '-z', start_I=body.Pb_I +
                     body.Rb_I @ Mat([0, -body.length/2, 0]))
    body.add_hookes_joint(L_thigh, about='xy')
    add_torque(body, L_thigh, about='xy')

    L_calf = Link3D('LL', '-z', start_I=L_thigh.bottom_I)
    add_foot(L_calf, at='bottom', nsides=8, friction_coeff=1.)
    L_thigh.add_revolute_joint(L_calf, about='y')
    add_torque(L_thigh, L_calf, about='y')

    # upper right, lower right
    R_thigh = Link3D('UR', '-z', start_I=body.Pb_I +
                     body.Rb_I @ Mat([0, body.length/2, 0]))
    body.add_hookes_joint(R_thigh, about='xy')
    add_torque(body, R_thigh, about='xy')

    R_calf = Link3D('LR', '-z', start_I=R_thigh.bottom_I)
    add_foot(R_calf, at='bottom', nsides=8, friction_coeff=1.)
    R_thigh.add_revolute_joint(R_calf, about='y')
    add_torque(R_thigh, R_calf, about='y')

    # combine into a robot
    robot = System3D(
        '3D biped', [body, tail, L_thigh, L_calf, R_thigh, R_calf])

    return robot, add_pyomo_constraints


def add_pyomo_constraints(robot: System3D):
    assert robot.m is not None,\
        'robot does not have a pyomo model defined on it'

    from math import pi as π
    body, tail, L_thigh, L_calf, R_thigh, R_calf = [
        link['q'] for link in robot.links]

    # tail
    constrain_rel_angle(robot.m, 'tail_pitch',
                        -π/2, body[:, :, 'theta'], tail[:, :, 'theta'], π/2)

    constrain_rel_angle(robot.m, 'tail_taw',
                        -π/2, body[:, :, 'psi'], tail[:, :, 'psi'], π/2)

    # left leg
    constrain_rel_angle(robot.m, 'left_hip',
                        -π/2, body[:, :, 'theta'], L_thigh[:, :, 'theta'], π/2)

    constrain_rel_angle(robot.m, 'left_knee',
                        0, L_thigh[:, :, 'theta'], L_calf[:, :, 'theta'], π)

    # right leg
    constrain_rel_angle(robot.m, 'right_hip',
                        -π/2, body[:, :, 'theta'], R_thigh[:, :, 'theta'], π/2)

    constrain_rel_angle(robot.m, 'right_knee',
                        0, R_thigh[:, :, 'theta'], R_calf[:, :, 'theta'], π)



from typing import Tuple, List, Iterable, TYPE_CHECKING
from pyomo.environ import Var, ConcreteModel
from .utils import flatten


if TYPE_CHECKING:
    from .links import Link3D


class VariableList:
    """Access pyomo variables using
    ```
    >>> variable_list[fe,cp]
    ```
    with fe = 1..nfe, cp = 1..ncp (ie, one-based)
    """
    # sources_of_vars is an iterable of Links. Typing is tricky due to circular imports

    def __init__(self, m: ConcreteModel, sources_of_vars: Iterable['Link3D']):
        self.sources_of_vars = sources_of_vars

        self._var_list: List[List[List[Var]]] = [
            [self._get_vars(fe, cp) for cp in m.cp] for fe in m.fe
        ]

    def _get_vars(self, fe: int, cp: int) -> List[Var]:
        return flatten(src.get_pyomo_vars(fe, cp) for src in self.sources_of_vars)

    def __getitem__(self, idx: Tuple[int, int]) -> List[Var]:
        fe, cp = idx
        assert (1 <= fe <= len(self._var_list)
                and 1 <= cp <= len(self._var_list[0]))

        return self._var_list[fe-1][cp-1]
